import pytest
import numpy as np

import huracanpy
from huracanpy._data._netcdf import _find_trajectory_id


def test_load_track():
    data = huracanpy.load(huracanpy.example_TRACK_file, tracker="TRACK")
    assert len(data.groupby("track_id")) == 2


def test_load_csv():
    data = huracanpy.load(huracanpy.example_csv_file, tracker="tempestextremes")
    assert len(data) == 13
    assert len(data.time) == 99
    assert len(data.groupby("track_id")) == 3


def test_load_netcdf():
    data = huracanpy.load(huracanpy.example_TRACK_netcdf_file)
    assert len(data.time) == 4580
    track_id = _find_trajectory_id(data)
    assert len(track_id) == 4580
    assert len(np.unique(track_id)) == 86


@pytest.mark.parametrize(
    "filename,tracker",
    [
        (huracanpy.example_TRACK_file, "TRACK"),
        (huracanpy.example_TRACK_netcdf_file, None),
        (huracanpy.example_csv_file, None),
    ],
)
def test_save_netcdf(filename, tracker, tmp_path):
    data = huracanpy.load(filename, tracker=tracker)
    # Copy the data because save modifies the dataset at the moment
    huracanpy.save(data.copy(), str(tmp_path / "tmp_file.nc"))

    # Reload the data and check it is still the same
    data_ = huracanpy.load(str(tmp_path / "tmp_file.nc"))

    for var in data_.variables:
        # Work around for xarray inconsistent loading the data as float or double
        # depending on fill_value and scale_factor
        # np.testing.assert_allclose doesn't work for datetime64
        if np.issubdtype(data[var].dtype, np.datetime64):
            assert (data[var].data == data_[var].data).all()
        elif data[var].dtype != data_[var].dtype:
            np.testing.assert_allclose(
                data[var].data.astype(data_[var].dtype), data_[var].data, rtol=1e-6
            )
        else:
            np.testing.assert_allclose(data[var].data, data_[var].data, rtol=0)


@pytest.mark.parametrize(
    "filename,tracker",
    [
        (huracanpy.example_TRACK_file, "TRACK"),
        (huracanpy.example_csv_file, None),
    ],
)
def test_save_csv(filename, tracker, tmp_path):
    data = huracanpy.load(filename, tracker=tracker)
    # Copy the data because save modifies the dataset at the moment
    huracanpy.save(data.copy(), str(tmp_path / "tmp_file.csv"))

    # Reload the data and check it is still the same
    data_ = huracanpy.load(str(tmp_path / "tmp_file.csv"))

    for var in data_.variables:
        # Work around for xarray inconsistent loading the data as float or double
        # depending on fill_value and scale_factor
        # np.testing.assert_allclose doesn't work for datetime64
        if np.issubdtype(data[var].dtype, np.datetime64):
            assert (data[var].data == data_[var].data).all()
        elif data[var].dtype != data_[var].dtype:
            np.testing.assert_allclose(
                data[var].data.astype(data_[var].dtype), data_[var].data, rtol=1e-6
            )
        else:
            np.testing.assert_allclose(data[var].data, data_[var].data, rtol=0)


@pytest.mark.parametrize(
    "subset,length",
    [
        ("wmo", 8),
        ("usa", 10),
    ],
)
def test_ibtracs_offline(subset, length):
    ib = huracanpy.load(tracker="ibtracs", ibtracs_subset=subset)
    assert ib.season.min() == 1980
    assert (
        len(ib.time) > 0
    )  # Can't assert on dataset length, because it might change with updates.
    assert (len(ib)) == length



"""
Utils related to storm category
"""

import numpy as np
import pint
import xarray as xr
import pandas as pd

from metpy.xarray import preprocess_and_wrap
from metpy.units import units


@preprocess_and_wrap(wrap_like="variable")
def categorize(variable, bins=None, labels=None, convention=None, variable_units=None):
    """Calculate a generic category from a variable and a set of thresholds

    Parameters
    ----------
    variable : array_like
        The variable to be categorized
    bins : array_like
        Bounds for the categories, including upper and lower bounds
    labels : array_like
        Name of the categories. len(labels) = len(bins) -1
    convention : str
        * Klotzbach
        * Simpson
        * Saffir-Simpson
    variable_units : str
        The units of the input variable. Only needs to be specified if they are
        different to the units of the bins and they are not already in the attributes of
        the variable.

    Returns
    -------
    numpy.ndarray
        The category label for each value in the input variable

    """
    if bins is None and labels is None:
        if convention is None:
            raise ValueError("Must specify either bins/labels or convention")
        else:
            bins = _thresholds[convention]["bins"]
            labels = _thresholds[convention]["labels"]

    if not isinstance(variable, pint.Quantity) or variable.unitless:
        if variable_units is None and isinstance(bins, pint.Quantity):
            variable_units = str(bins.units)
        variable = variable * units(variable_units)

    categories = np.zeros_like(variable) * np.nan
    for n, label in enumerate(labels):
        categories[(bins[n] < variable) & (variable <= bins[n + 1])] = label

    return categories


_thresholds = {
    "Klotzbach": dict(
        bins=np.array([-np.inf, 925, 945, 960, 975, 990, 1005, np.inf]) * units("hPa"),
        labels=[5, 4, 3, 2, 1, 0, -1],
    ),
    "Simpson": dict(
        bins=np.array([-np.inf, 920, 945, 965, 970, 980, 990, np.inf]) * units("hPa"),
        labels=[5, 4, 3, 2, 1, 0, -1],
    ),
    "Saffir-Simpson": dict(
        bins=np.array([-np.inf, 16, 29, 38, 44, 52, 63, np.inf]) * units("m s-1"),
        labels=[-1, 0, 1, 2, 3, 4, 5],
    ),
}


@preprocess_and_wrap(wrap_like="wind")
def get_sshs_cat(wind, convention="Saffir-Simpson", wind_units="m s-1"):
    """
    Function to determine the Saffir-Simpson Hurricane Scale (SSHS) category.

    Parameters
    ----------
    wind : xarray.DataArray
        10-minutes averaged 10m wind in m/s

    wind_units : str, default="m s-1"
        The units of the input array if they are not already provided by the attributes

    Returns
    -------
    xarray.DataArray
        The category series.
        You can append it to your tracks by running tracks["sshs"] = get_sshs_cat(tracks.wind)
    """
    return categorize(wind, convention=convention, variable_units=wind_units)


@preprocess_and_wrap(wrap_like="slp")
def get_pressure_cat(slp, convention="Klotzbach", slp_units="hPa"):
    """
    Determine the pressure category according to selected convention.

    Parameters
    ----------
    slp : xarray.DataArray
        Minimum Sea-level Pressure series in hPa
    convention : str
        Name of the classification convention you want to use.
            * Klotzbach (default)
            * Simpson
    slp_units : str, default="hPa"
        The units of the input array if they are not already provided by the attributes

    Returns
    -------
    xarray.DataArray
        The category series.
        You can append it to your tracks by running tracks["cat"] = get_pressure_cat(tracks.slp)

    """
    if not isinstance(slp, pint.Quantity) or slp.unitless:
        if slp.min() > 10000 and slp_units == "hPa":
            print(
                "Caution, pressure are likely in Pa, they are being converted to hPa "
                "for categorization. In future specify the units explicitly by passing "
                'slp_units="Pa" to this function or setting '
                'slp.attrs["units"] = "Pa"'
            )
            slp = slp / 100

    return categorize(slp, convention=convention, variable_units=slp_units)


# [Stella] Leaving that here as an alternative method memo if we encounter performance issues.
def categorize_alt(var, bins, labels=None):
    """
    Provides category according to provided bins and labels

    Parameters
    ----------
    var : xarray.DataArray
        The variable to categorize
    bins : array_like
        bins boundaries
    labels : array_like, optional
        Name of the categories. len(labels) = len(bins) -1

    Returns
    -------
    xarray.DataArray
        The category series.
        You can append it to your tracks by running tracks["cat"] = categorize(tracks.var, bins)
    """
    cat = pd.cut(var, bins, labels=labels)
    return xr.DataArray(cat, dims="record", coords={"record": var.record})



"""Huracanpy module for tracks diagnostics"""

__all__ = ["track_density", "track_stats", "translation_speed", "lifecycle"]

from . import track_density
from . import track_stats
from . import translation_speed
from . import lifecycle



import huracanpy


def test_trackswhere():
    tracks = huracanpy.load(huracanpy.example_csv_file)

    tracks["category"] = huracanpy.utils.category.get_pressure_cat(
        tracks.slp, slp_units="Pa"
    )

    tracks_subset = huracanpy.subset.trackswhere(
        tracks, lambda track: track.category.max() >= 2
    )

    assert set(tracks_subset.track_id.data) == {0, 2}



"""Matching functions"""

import pandas as pd
import numpy as np
from haversine import haversine_vector, Unit
from itertools import combinations


def match_pair(
    tracks1,
    tracks2,
    name1="1",
    name2="2",
    max_dist=300,
    min_overlap=0,
):
    """

    Parameters
    ----------
    tracks1 (pd.DataFrame): the first track set to match
    tracks2 (pd.DataFrame): the second tracks set to match
    name1 (str): Suffix for the first dataframe
    name2 (str): Suffix for the second dataframe
    max_dist (float) : Threshold for maximum distance between two tracks
    min_overlap (int) : Minimum number of overlapping time steps for matching

    Returns
    -------
    pd.DataFrame
        Dataframe containing the matching tracks with
            the id from both datasets
            the number of matching time steps
            the distance between two tracks
    """

    # Prepare data
    tracks1, tracks2 = (
        tracks1[["track_id", "lon", "lat", "time"]].to_dataframe(),
        tracks2[["track_id", "lon", "lat", "time"]].to_dataframe(),
    )
    tracks1["lon"] = np.where(tracks1.lon > 180, tracks1.lon - 360, tracks1.lon)
    tracks2["lon"] = np.where(tracks2.lon > 180, tracks2.lon - 360, tracks2.lon)

    # Find corresponding points (same time step, less than max_dist km)
    merged = pd.merge(tracks1, tracks2, on="time")
    X = np.concatenate([[merged.lat_x], [merged.lon_x]]).T
    Y = np.concatenate([[merged.lat_y], [merged.lon_y]]).T
    merged["dist"] = haversine_vector(X, Y, unit=Unit.KILOMETERS)
    merged = merged[merged.dist <= max_dist]
    # Compute temporal overlap
    temp = (
        merged.groupby(["track_id_x", "track_id_y"])[["dist"]]
        .count()
        .rename(columns={"dist": "temp"})
    )
    # Build a table of all pairs of tracks sharing at least one point
    matches = (
        merged[["track_id_x", "track_id_y"]]
        .drop_duplicates()
        .join(temp, on=["track_id_x", "track_id_y"])
    )
    matches = matches[matches.temp >= min_overlap]
    dist = merged.groupby(["track_id_x", "track_id_y"])[["dist"]].mean()
    matches = matches.merge(dist, on=["track_id_x", "track_id_y"])
    # Rename columns before output
    matches = matches.rename(
        columns={"track_id_x": "id_" + name1, "track_id_y": "id_" + name2}
    )
    return matches


def match_multiple(
    datasets,
    names,
    max_dist=300,
    min_overlap=0,
):
    """
    Function to match any number of tracks sets

    Parameters
    ----------
    datasets : list of xr.Dataset
        list of the sets to be matched.
    names : list of str
        labels for the datasets. names must have the same length as datasets
    max_dist : float
        Threshold for maximum distance between two tracks
    min_overlap : int
        Minimum number of overlapping time steps for matching

    Raises
    ------
    NotImplementedError
        If two datasets have no match.

    Returns
    -------
    M : pd.dataframe
        table of matching tracks among all the datasets

    """

    assert len(datasets) == len(names), "datasets and names must have the same length."

    M = pd.DataFrame(columns=["id_" + n for n in names[:2]])
    for names_pair, dataset_pair in zip(
        combinations(names, 2), combinations(datasets, 2)
    ):
        m = match_pair(*dataset_pair, *names_pair, max_dist, min_overlap)
        if len(m) == 0:
            raise NotImplementedError(
                "For the moment, the case where two datasets have no match is not handled. Problem raised by datasets "  # TODO
                + str(names_pair)
            )
        M = M.merge(m[["id_" + names_pair[0], "id_" + names_pair[1]]], how="outer")
    return M


# TODO: Deal with duplicates: merge, max...?



"""Huracanpy module for useful auxilliary functions"""

__all__ = ["geography", "category", "time", "add_all_info"]

from . import geography
from . import category
from . import time


def add_all_info(
    data,
    lat_name="lat",
    lon_name="lon",
    track_id_name="track_id",
    time_name="time",
    year_name="year",
    month_name="month",
    day_name="day",
    hour_name="hour",
    wind_name="wind10",
    wind_units="m s-1",
    slp_name="slp",
    slp_units="hPa",
    sshs_convention="Saffir-Simpson",
    pres_cat_convention="Klotzbach",
):
    """


    Parameters
    ----------
    data : the tracks xarray object
    lat_name : str, optional
        Name of the latitude variable. The default is "lat".
    lon_name : str, optional
        Name of the longitude variable. The default is "lon".
    track_id_name : str, optional
        Name of the track ID variable.. The default is "track_id".
    time_name : str, optional
        Name of the time variable. If set to None, no time-related attributes are computed. The default is "time".
    year_name : str, optional
        Name of the year variable. The default is "year".
    month_name : str, optional
        Name of the month variable. The default is "month".
    day_name : str, optional
        Name of the day variable. The default is "day".
    hour_name : str, optional
        Name of the hour variable. The default is "hour".
    wind_name : str, optional
        Name of the winf variable. The default is "wind10".
    wind_units : str, optional
        The default is "m s-1".
    slp_name : str, optional
        Name of the SLP variable. The default is "slp".
    slp_units : str, optional
        The default is "hPa".
    sshs_convention : str, optional
        The default is "Saffir-Simpson".
    pres_cat_convention : str, optional
        The default is "Klotzbach".

    Returns
    -------
    data : xr.DataSet
        Input dataset with additional info.

    """

    # Geographical
    if lat_name is not None:
        data["hemisphere"] = geography.get_hemisphere(data[lat_name])
        if lon_name is not None:
            data["basin"] = geography.get_basin(data[lon_name], data[lat_name])
            data["is_ocean"] = geography.get_land_or_ocean(
                data[lon_name], data[lat_name]
            )
            data["country"] = geography.get_country(data[lon_name], data[lat_name])
            data["continent"] = geography.get_continent(data[lon_name], data[lat_name])

    # Time
    if time_name is not None:
        if time_name not in list(data.keys()):
            data[time_name] = time.get_time(
                data[year_name],
                data[month_name],
                data[day_name],
                data[hour_name],
            )
        data["season"] = time.get_season(
            data[track_id_name], data[lat_name], data[time_name]
        )

    # Category
    if wind_name is not None:
        data["sshs"] = category.get_sshs_cat(
            data[wind_name], sshs_convention, wind_units
        )
    if slp_name is not None:
        data["pres_cat"] = category.get_pressure_cat(
            data[slp_name], pres_cat_convention, slp_units
        )

    return data



"""
Module to load tracks stored as csv files, including TempestExtremes output.
"""

import pandas as pd


from .. import utils


def load(
    filename,
    read_csv_kws=dict(),
):
    """Load csv tracks data as an xarray.Dataset
    These tracks may come from TempestExtremes StitchNodes, or any other source.

    Parameters
    ----------
    filename : str
        The file must contain at least longitude, latitude, time and track ID.
            - longitude and latitude can be named that, or lon and lat.
            - time must be defined a single `time`column or by four columns : year, month, day, hour
            - track ID must be within a column named track_id.

    Returns
    -------
    xarray.Dataset
    """

    ## Read file
    tracks = pd.read_csv(filename, **read_csv_kws)
    if (
        tracks.columns.str[0][1] == " "
    ):  # Sometimes columns names are read starting with a space, which we remove
        tracks = tracks.rename(columns={c: c[1:] for c in tracks.columns[1:]})
    tracks.columns = tracks.columns.str.lower()  # Make all column names lowercase
    tracks = tracks.rename(
        {"longitude": "lon", "latitude": "lat"}
    )  # Rename lon & lat columns if necessary

    ## Geographical attributes
    if "lon" in tracks.columns:
        tracks.loc[tracks.lon < 0, "lon"] += (
            360  # Longitude are converted to [0,360] if necessary
        )

    ## Time attribute
    if "iso_time" in tracks.columns:
        tracks["time"] = pd.to_datetime(tracks.iso_time)
        tracks = tracks.drop(columns="iso_time")
    elif "time" in tracks.columns:
        tracks["time"] = pd.to_datetime(tracks.time)
    else:
        tracks["time"] = utils.time.get_time(
            tracks.year, tracks.month, tracks.day, tracks.hour
        )

    # Output xr dataset
    return tracks.to_xarray().rename({"index": "record"})



import xarray as xr


def trackswhere(tracks, condition):
    """Subset tracks from the input

    e.g select all tracks that are solely in the Northern hemisphere
    >>> tracks_subset = huracanpy.subset.trackswhere(tracks, lambda x: (x.lat > 0).all())

    Parameters
    ----------
    tracks : xarray.Dataset
    condition : function
        A function that takes an `xarray.Dataset` of an individual track and returns
        True or False

    Returns
    -------
    xarray.Dataset
        A dataset with the subset of tracks from the input that match the given criteria

    """
    track_groups = tracks.groupby("track_id")

    if callable(condition):
        is_match = track_groups.map(condition)

    track_groups = [
        track for n, (track_id, track) in enumerate(track_groups) if is_match[n]
    ]

    assert len(tracks.time.dims) == 1
    return xr.concat(track_groups, dim=tracks.time.dims[0])



"""Functions for score computation"""

import numpy as np


def POD(matches, ref, ref_name="ib"):
    """

    Parameters
    ----------
    matches
    ref
    ref_name

    Returns
    -------

    """
    N_detected = matches["id_" + ref_name].nunique()
    N_total = len(np.unique(ref.track_id.values))
    return N_detected / N_total


def FAR(matches, detected, detected_name="UZ"):
    """

    Parameters
    ----------
    matches
    detected
    detected_name

    Returns
    -------

    """
    N_matched = matches["id_" + detected_name].nunique()
    N_total = len(np.unique(detected.track_id.values))
    return 1 - (N_matched / N_total)



"""
Module containing functions to compute lifecycle stage
"""


def time_from_genesis(data):
    """
    Output the time since genesis for each TC point

    Parameters
    ----------
    data : TYPE
        DESCRIPTION.

    Returns
    -------
    xarray.DataArray
        The time_from_genesis series.
        You can append it to your tracks by running tracks["time_from_genesis"] = time_from_genesis(tracks)

    """
    data_df = data[["track_id", "time"]].to_dataframe()
    data_df = data_df.merge(
        data_df.groupby("track_id").time.min(),
        on="track_id",
        suffixes=["_actual", "_gen"],
    )
    time_from_start = data_df.time_actual - data_df.time_gen
    return (
        time_from_start.to_xarray().rename({"index": "obs"}).rename("time_from_genesis")
    )


def time_from_extremum(data, varname, stat=max):
    """

    Parameters
    ----------
    data
    varname
    stat

    Returns
    -------

    """
    if stat == "max":
        asc = False
    elif stat == "min":
        asc = True
    else:
        raise NotImplementedError("stat not recognized. Please use one of {min, max}")
    data_df = data[["track_id", "time", varname]].to_dataframe()
    extr = data_df.sort_values(varname, ascending=asc).groupby("track_id").first()
    data_df = data_df.merge(extr, on="track_id", suffixes=["_actual", "_extr"])
    time_from_extr = data_df.time_actual - data_df.time_extr
    return (
        time_from_extr.to_xarray().rename({"index": "obs"}).rename("time_from_extremum")
    )



"""
Functions to plot the tracks
"""

import matplotlib.pyplot as plt
import seaborn as sns
import cartopy.crs as ccrs


def plot_tracks_basic(
    tracks,
    intensity_var=None,
    subplot_kws=dict(projection=ccrs.PlateCarree(180)),
    fig_kws=dict(figsize=(10, 10)),
    scatter_kws=dict(palette="nipy_spectral", s=2),
):
    assert "lon" in list(tracks.keys()), "lon is not present in the data"
    assert "lat" in list(tracks.keys()), "lat is not present in the data"

    fig, ax = plt.subplots(subplot_kw=subplot_kws, **fig_kws)
    ax.coastlines()
    sns.scatterplot(
        data=tracks,
        x="lon",
        y="lat",
        hue=intensity_var,
        ax=ax,
        **scatter_kws,
        transform=ccrs.PlateCarree(),
    )

    return fig, ax



"""Huracanpy module for plots"""

__all__ = ["plot_tracks_basic", "plot_density", "fancyline", "venn"]

from ._fancyline import fancyline
from .tracks import plot_tracks_basic
from .density import plot_density
from .venn import venn



"""
Functions to plot track/genesis/whatever density

To compute the density, see huracanpy.diags.track_density
"""

import matplotlib.pyplot as plt
import cartopy.crs as ccrs


def plot_density(
    d,
    contourf_kws=dict(cmap="magma_r", levels=10),
    subplot_kws=dict(projection=ccrs.PlateCarree(180)),
    fig_kws=dict(),
):
    fig, ax = plt.subplots(subplot_kw=subplot_kws, **fig_kws)
    ax.coastlines()
    d.where(d > 0).plot.contourf(ax=ax, transform=ccrs.PlateCarree(), **contourf_kws)

    return fig, ax



"""
Module containing functions to compute translation speed
"""

from haversine import haversine
import numpy as np
import xarray as xr


def translation_speed(data):
    """
    Compute translation speed along tracks

    Parameters
    ----------
    data : xarray.Dataset

    Returns
    -------
    xarray.Dataset
        Translation speeds. Output is stored for points that correspond to the middle of two consecutive points in the initial dataset.

    """
    data = data.sortby(["track_id", "time"])
    V, lat, lon, t, tid = [], [], [], [], []  # To store results temporarily

    dims = data.time.dims
    assert len(dims) == 1
    dim = dims[0]
    for i in range(len(data[dim]) - 1):
        p = data.isel(**{dim: i})  # Current point
        q = data.isel(**{dim: i + 1})  # Next point
        if p.track_id == q.track_id:  # If both points belong to the same track
            dt = np.timedelta64((q.time - p.time).values, "s")  # Temporal interval in s
            dx = haversine(
                (p.lat.values[()], p.lon.values[()]),
                (q.lat.values[()], q.lon.values[()]),
                unit="m",
            )  # Displacement in m
            v = dx / dt.astype(float)  # translation speed in m/s
            V.append(v)
            # Results will be stored with coordinates corresponding to the middle of p and q
            lat.append((p.lat + q.lat).values / 2)
            lon.append((p.lon + q.lon).values / 2)
            t.append((p.time + (q.time - p.time) / 2).values)
            tid.append(p.track_id.values)

    # Transform into clean dataset
    V = xr.DataArray(V, dims="mid_record", coords={"mid_record": np.arange(len(V))})
    lon = xr.DataArray(
        lon, dims="mid_record", coords={"mid_record": np.arange(len(lon))}
    )
    lat = xr.DataArray(
        lat, dims="mid_record", coords={"mid_record": np.arange(len(lat))}
    )
    t = xr.DataArray(t, dims="mid_record", coords={"mid_record": np.arange(len(t))})
    tid = xr.DataArray(
        tid, dims="mid_record", coords={"mid_record": np.arange(len(tid))}
    )

    return xr.Dataset(
        {"lon": lon, "lat": lat, "time": t, "track_id": tid, "translation_speed": V}
    )



import gzip

import datetime
import cftime
import numpy as np
import xarray as xr

from parse import parse

# Use align specifications (^, <, >) to allow variable whitespace in headers
# Left aligned (<) for "nvars" so nfields takes all whitespace between in case there is
# only one space
# Right aligned (>) for variables at the end since lines are stripped of whitespace
# prior to parsing
header_fmt = "TRACK_NUM{ntracks:^d}ADD_FLD{nfields:^d}{nvars:<d}&{var_has_coords}"
track_header_fmt = "TRACK_ID{track_id:>d}"
track_header_fmt_new = "TRACK_ID{track_id:^d}START_TIME{start_time:>}"
track_info_fmt = "POINT_NUM{npoints:>d}"


def _parse(fmt, string, **kwargs):
    # Call parse but raise an error if None is returned
    result = parse(fmt, string, **kwargs)

    if result is None:
        raise ValueError(f"Format {fmt} does not match string {string}")

    return result


def parse_date(date, calendar=None):
    if len(date) == 10:  # i.e., YYYYMMDDHH
        if calendar is not None:
            yr = int(date[0:4])
            mn = int(date[4:6])
            dy = int(date[6:8])
            hr = int(date[8:10])
            return cftime.datetime(yr, mn, dy, hr, calendar=calendar)
        else:
            return datetime.datetime.strptime(date.strip(), "%Y%m%d%H")
    else:
        return int(date)


def load(filename, calendar=None, variable_names=None):
    """Load ASCII TRACK data as an xarray.Dataset

    Parameters
    ----------
    filename: str
    calendar : optional
    variable_names : list of str, optional
        TRACK

    Returns
    -------
    xarray.Dataset
    """
    if filename.split(".")[-1] == "gz":
        open_func = gzip.open
    else:
        open_func = open

    output = list()
    with open_func(filename, "rt") as f:
        # The first lines can contain extra information bounded by two extra lines
        # Just skip to the main header line for now
        line = ""
        while not line.startswith("TRACK_NUM"):
            line = f.readline().strip()

        # Load information about tracks from header line
        # If there are no added variables the line ends at the "&"
        try:
            header = _parse(header_fmt, line).named
        except ValueError:
            header = _parse(header_fmt.split("&")[0] + "&", line).named
            header["var_has_coords"] = ""

        ntracks = header["ntracks"]
        nfields = header["nfields"]
        nvars = header["nvars"]
        has_coords = [int(x) == 1 for x in header["var_has_coords"]]

        # Check header data is consistent
        assert len(has_coords) == nfields
        assert sum([3 if x == 1 else 1 for x in has_coords]) == nvars

        # Create a list of variables stored in each track
        # Generic names for variables as there is currently no information otherwise
        var_labels = ["lon", "lat", "vorticity"]
        if variable_names is None:
            variable_names = [f"feature_{n}" for n in range(nfields)]
        else:
            assert len(variable_names) == nfields
        for n, variable_name in enumerate(variable_names):
            if has_coords[n]:
                var_labels.append(f"{variable_name}_longitude")
                var_labels.append(f"{variable_name}_latitude")
            var_labels.append(variable_name)

        # Read in each track as an xarray dataset with time as the coordinate
        for n in range(ntracks):
            # Read individual track header (two lines)
            line = f.readline().strip()
            try:
                track_info = _parse(track_header_fmt, line).named
            except ValueError:
                track_info = _parse(track_header_fmt_new, line).named

            line = f.readline().strip()
            npoints = _parse(track_info_fmt, line)["npoints"]

            # Generate arrays for time coordinate and variables
            # Time is a list because it will hold datetime or cftime objects
            # Other variables are a dictionary mapping variable name to a tuple of
            # (time, data_array) as this is what is passed to xarray.Dataset
            times = [None] * npoints
            track_data = {label: ("record", np.zeros(npoints)) for label in var_labels}

            # Add track ID as a variable along the record dimension so that it can be used
            # for groupby
            track_id = np.array([track_info["track_id"]] * npoints)
            track_data["track_id"] = ("record", track_id)

            # Populate time and data line by line
            for m in range(npoints):
                line = f.readline().strip().split("&")
                time, lon, lat, vorticity = line[0].split()
                times[m] = parse_date(time, calendar=calendar)
                track_data["lon"][1][m] = float(lon)
                track_data["lat"][1][m] = float(lat)
                track_data["vorticity"][1][m] = float(vorticity)

                for i, label in enumerate(var_labels[3:], start=1):
                    track_data[label][1][m] = float(line[i])

            # Return a dataset for the individual track
            # Add time separately so xarray can deal with the awkward np.datetime64
            # format
            ds = xr.Dataset(track_data)
            ds["time"] = ("record", times)
            output.append(ds)

    output = xr.concat(output, dim="record")
    output.track_id.attrs["cf_role"] = "trajectory_id"

    return output



"""
Module containing function to compute track densities
"""

import numpy as np
import xarray as xr


def simple_global_histogram(lon, lat, bin_size=5, N_seasons=1):
    """
    Function to compute the track density, based on a simple 2D histogram.


    Parameters
    ----------
    lon : xarray.DataArray
        longitude series
    lat : xarray.DataArray
        latitude series
    bin_size : int or float, optional
        The size in degree of the bins. The default is 5.
    N_seasons : int, optional
        Number of season (will be used to divide the final results, so that is represents points/year). The default is 1.

    Returns
    -------
    xarray.DataArray
        Histogram representing number of point per bin per season.

    """

    # compute 2D histogram
    x = np.arange(0, 360 + bin_size, bin_size)
    y = np.arange(-90, 90 + bin_size, bin_size)
    H, X, Y = np.histogram2d(lon, lat, bins=[x, y])
    # Turn into xarray
    da = xr.DataArray(
        H,
        dims=["lon", "lat"],
        coords={"lon": (X[:-1] + X[1:]) / 2, "lat": (Y[:-1] + Y[1:]) / 2},
    )
    # Format
    da = da.where(da > 0).transpose() / N_seasons
    return da



"""huracanpy - A python package for working with various forms of feature tracking data"""

__version__ = "0.1.0"
__author__ = "Leo Saffin <l.saffin@reading.ac.uk>, Stella Bourdin <stella.bourdin@physics.ox.ac.uk>, Kelvin Ng "
__all__ = [
    "load",
    "save",
    "utils",
    "diags",
    "plot",
    "assess",
    "subset",
    "example_csv_file",
    "example_year_file",
    "example_TRACK_file",
    "example_TRACK_netcdf_file",
]


from . import utils
from ._data import (
    load,
    save,
    example_csv_file,
    example_year_file,
    example_TRACK_file,
    example_TRACK_netcdf_file,
)
from . import diags
from . import plot
from . import assess
from . import subset



"""Huracanpy module for tracker assessment"""

__all__ = ["match_pair", "match_multiple", "scores", "overlap"]

from ._match import match_pair, match_multiple
from . import scores
from .overlap import overlap



from . import _csv
from . import _TRACK
from . import _netcdf
from . import ibtracs
from huracanpy import utils


def load(
    filename=None,
    tracker=None,
    add_info=False,
    ibtracs_online=False,
    ibtracs_subset="wmo",
    ibtracs_clean=True,
    **kwargs,
):
    """Load track data

    The optional parameters for different trackers (currently **IBTrACS** and **TRACK**)
    are named {tracker}_{parameter} (in lower case), e.g. "ibtracs_online".

    Parameters
    ----------
    filename : str, optional
        The file to be loaded. If `tracker="ibtracs"`, this is not needed as the data is
        either included in huracanpy or downloaded when called
    tracker : str, optional
        If the file is not a CSV or NetCDF (identified by the file extension) then the
        tracker needs to be specified to decide how to load the data

        * **track**
        * **csv**, **te**, **tempestextremes**, **uz**:
        * **ibtracs**

    add_info : bool, default=False
    ibtracs_online : bool, default=False
        * **False**: Use a small subset of the IBTrACS data included in this package
        * **True**: Download the IBTrACS data
    ibtracs_subset : str, default="ALL"
        IBTrACS subset. When loading offline data it is one of

        * **WMO**: Data with the wmo_* variables. The data as reported by the WMO agency
          responsible for each basin, so methods are not consistent across basins
        * **USA** or **JTWC**: Data with the usa_* variables. The data as recorded by
          the USA/Joint Typhoon Warning Centre. Methods are consistent across basins,
          but may not be complete.

        If you are downloading the online data, the subsets are the different files
        provided by IBTrACS

        * **ACTIVE**: TCs currently active
        * **ALL**: Entire IBTrACS database
        * Specific basins: **EP**, **NA**, **NI**, **SA**, **SI**, **SP**, **WP**
        * **last3years**: self-explanatory
        * **since1980**: Entire IBTrACS database since 1980 (advent of satellite era,
          considered reliable from then on)

    ibtracs_clean : bool, default=True
        If downloading IBTrACS data, this parameter says whether to delete the
        downloaded file after loading it into memory.
    **kwargs
        When loading tracks from a netCDF file, these keyword arguments will be passed
        to :func:`xarray.open_dataset`

        When loading tracks from a TRACK ASCII file, the following arguments can be used

        * variable_names : list of str, optional

          When loading data from a TRACK ASCII file, specify the list of variables that
          have been added to the tracks. This does not include time/lon/lat/vorticity
          which are in the track by default. If vorticity at multiple levels has been
          added to the tracks, then a variable name for each level needs to be included.
          If the variable names are not given, then any additional variables will be
          named variable_n, where n goes from 0 to the number of variables
        * track_calendar : str, optional

          When loading data from a TRACK ASCII file, if the data uses a different calendar
          to the default :class:`datetime.datetime`, then you can pass this argument to
          load the times in as :class:`cftime.datetime` with the given calendar instead

    Returns
    -------
    xarray.Dataset

    """
    # If tracker is not given, try to derive the right function from the file extension
    if tracker is None:
        if filename.split(".")[-1] == "csv":
            data = _csv.load(filename)
        elif filename.split(".")[-1] == "nc":
            data = _netcdf.load(filename, **kwargs)
        else:
            raise ValueError(f"{tracker} is set to None and file type is not detected")

    # If tracker is given, use the relevant function
    else:
        if tracker.lower() == "track":
            data = _TRACK.load(filename, **kwargs)
        elif tracker.lower() in ["csv", "te", "tempestextremes", "uz"]:
            data = _csv.load(filename)
        elif tracker.lower() == "ibtracs":
            if ibtracs_online:
                if filename is None:
                    filename = "ibtracs.csv"

                with ibtracs.online(ibtracs_subset, filename, ibtracs_clean) as f:
                    data = _csv.load(
                        f,
                        read_csv_kws=dict(
                            header=0,
                            skiprows=[1],
                            na_values=["", " "],
                            converters={
                                "SID": str,
                                "SEASON": int,
                                "BASIN": str,
                                "SUBBASIN": str,
                                "LON": float,
                                "LAT": float,
                            },
                        ),
                    )
            else:
                data = _csv.load(ibtracs.offline(ibtracs_subset))
        else:
            raise ValueError(f"Tracker {tracker} unsupported or misspelled")

    if add_info:  # TODO : Manage potentially different variable names
        data["hemisphere"] = utils.geography.get_hemisphere(data.lat)
        data["basin"] = utils.geography.get_basin(data.lon, data.lat)
        data["season"] = utils.time.get_season(data.track_id, data.lat, data.time)
        if "wind10" in list(data.keys()):  # If 'wind10' is in the attributes
            data["sshs"] = utils.category.get_sshs_cat(data.wind10)
        if "slp" in list(data.keys()):  # If 'slp' is in the attributes
            data["pres_cat"] = utils.category.get_pressure_cat(data.slp)

    return data



"""
Module containing functions to compute track statistics
"""


def duration(tracks):
    """
    Compute the duration of each track

    Parameters
    ----------
    tracks : xarray.Dataset

    Returns
    -------
    xarray.DataArray
        Duration of each track

    """
    duration = (
        tracks.groupby("track_id")
        .map(lambda x: x.time.max() - x.time.min())
        .rename("duration")
    )
    duration = (duration * 1e-9 / 3600).astype(float)
    duration.attrs["units"] = "h"
    return duration


def gen_vals(tracks):
    """
    Shows the attributes for the genesis point of each track

    Parameters
    ----------
    tracks : xarray.DataSet

    Returns
    -------
    xarray.Dataset
        Dataset containing only genesis points, with track_id as index.

    """

    return tracks.sortby("time").groupby("track_id").first()


def extremum_vals(tracks, varname, stat="max"):
    """
    Shows the attribute for the extremum point of each track

    Parameters
    ----------
    tracks : xarray.DataSet
    var : str
        The extremum variable
    stat : str, optional
        Type of extremum. Can be "min" or "max". The default is "max".

    Raises
    ------
    NotImplementedError
        If another value than "min" and "max" is given to stat

    Returns
    -------
    xarray.Dataset
        Dataset containing only extremum points, with track_id as index.

    """

    # tracks will be sorted along var and then the first line of each track_id will be used
    # asc determines whether the sorting must be ascending (True) or descending (False)
    if stat == "max":
        asc = False
    elif stat == "min":
        asc = True
    else:
        raise NotImplementedError("stat not recognized. Please use one of {min, max}")

    return tracks.sortby(varname, ascending=asc).groupby("track_id").first()



"""huracanpy module for shared data"""

__all__ = [
    "load",
    "save",
    "example_csv_file",
    "example_year_file",
    "example_TRACK_file",
    "example_TRACK_netcdf_file",
]

import pathlib

from ._load import load
from ._save import save

here = pathlib.Path(__file__).parent
testdata_dir = here / "example_data"

example_TRACK_file = str(
    testdata_dir / "tr_trs_pos.2day_addT63vor_addmslp_add925wind_add10mwind.tcident.new"
)

example_csv_file = str(testdata_dir / "sample.csv")

example_year_file = str(testdata_dir / "ERA5_1996_UZ.csv")

example_TRACK_netcdf_file = str(
    testdata_dir
    / "tr_trs_pos.2day_addT63vor_addmslp_add925wind_add10mwind.tcident.new.nc"
)
